#!/usr/bin/python3
import threading
import sys
import time
import os
import datetime
import zipfile
import subprocess
import logging
import signal
import getopt
import re
import shutil
import ctypes

# Disable QuickEdit mode by setting it's flag to 0x00
kernel32 = ctypes.WinDLL('kernel32', use_last_error=True)
kernel32.SetConsoleMode(kernel32.GetStdHandle(-10), (0x4 | 0x80 | 0x20 | 0x2 | 0x10 | 0x1 | 0x00 | 0x100))
del kernel32  # We don't need this variable floating around anymore, so we delete it

# Get the script parameters
try:
  opts, args = getopt.getopt(sys.argv[1:], 'hdmu', ['help', 'debug', 'memory', 'update'])
except getopt.GetoptError:
  print('Parsing of parameters failed. Running in default mode.')
  opts = list()

LOG_LEVEL = logging.INFO
print_memory_usage = False
for opt, arg in opts:
  if opt in ('-h', '--help'):
    print('Options:\n'
          '-d --debug: Run I with debug output.\n'
          '-h --help: Print help and exit.\n'
          '-m --memory: Uses guppy to print basic memory information once per minute. guppy must be installed manually!'
          '-u --update: Install libraries in requirements-{}.txt and update backup.py to newest version.'.format(sys.version_info[0]))
    sys.exit()
  elif opt in ('-d', '--debug'):
    LOG_LEVEL = logging.DEBUG
  elif opt in ('-m', '--memory'):
    print_memory_usage = True
  elif opt in ('-u', '--update'):
    libraries_updated = False
    try:  # Attempt to create/update pip.ini file
      loc = os.path.join(os.getenv('APPDATA'), 'pip')
      try:
        os.makedirs(loc)
      except:
        pass
      file_name = os.path.join(loc, 'pip.ini')
      if not os.path.exists(file_name):  # If there's no pip.ini file, make it.
        with open(file_name, 'w') as pip_ini:
          pip_ini.write(
            '[global]\n'
            'timeout=60\n'
            'trusted-host=172.16.15.21\n'
            'index-url=http://172.16.15.21:8080/simple')
    except:  # No big deal, if it fails; updating will just take longer.
      pass
    try:
      subprocess.check_call('{} -m pip install --index-url http://172.16.15.21:8080/simple/ --trusted-host 172.16.15.21 --upgrade pip'.format(sys.executable))  # We have to set default Python encoding for this to work
    except (subprocess.CalledProcessError, LookupError):
      os.execv(sys.executable, [sys.executable] + sys.argv)
    try:  # Update siwim_backup
      PACKAGE_NAME = 'siwim_backup'
      SCRIPT_FOLDER = 'D:/siwim_mkiii/backup_script'
      SCRIPT = os.path.join(SCRIPT_FOLDER, 'backup.py')

      output = subprocess.check_output('{} -m pip download --index-url http://172.16.15.21:8080/simple/ --trusted-host 172.16.15.21 siwim-backup'.format(sys.executable)).decode()
      archive = re.findall(r'{}-\d\.\d.\d\.zip'.format(PACKAGE_NAME), output)[0]

      file_path = '{}/{}/backup_win.py'.format(archive.rsplit('.', 1)[0], PACKAGE_NAME)
      try:
        os.makedirs(SCRIPT_FOLDER)
      except:  # We don't care if the file exists.
        pass
      with zipfile.ZipFile(archive, 'r') as zip:
        zip.extract(file_path)

      shutil.move(file_path, SCRIPT)
      os.remove(archive)
      shutil.rmtree(archive.rsplit('.', 1)[0])
    except Exception as e:
      print('Downloading siwim_backup failed: {}'.format(e))
    while not libraries_updated:  # In most cases this will only be run once, but we can't let the application run if updating libraries fails.
      try:
        subprocess.check_call([sys.executable, '-m', 'pip', 'install', '--index-url', 'http://172.16.15.21:8080/simple/', '--trusted-host', '172.16.15.21', '-r', 'requirements-{}.txt'.format(sys.version_info[0])])
        libraries_updated = True
      except Exception as e:  # If this fails for whatever reason just keep on going ...
        print('Updating of libraries failed: {}\nRetrying ...'.format(e))
    os.execv(sys.executable, [sys.executable] + [f for f in sys.argv if f not in ['-u', '--update']])  # Restart with previous flags except for update flag.

try:  # Make sure all required libraries are installed. Appropriate requirements.txt is required. If a library which isn't included in any of these paths is the only one missing, just add an import to it anyways.
  from lxml import etree
  from communication_module import CommunicationModule
  import requests
  from py_logging.cestel_logging import init_logger, configure_default_logger, log, log_exception
except Exception as e:
  print('Some libraries are missing. Try running with --update flag:{}\n'.format(e))
  time.sleep(5)
  sys.exit(1)

if os.path.dirname(sys.argv[0]) != '':
  os.chdir(os.path.dirname(sys.argv[0]))
configure_default_logger(level=LOG_LEVEL, console=LOG_LEVEL)

# This is the logger that handles siwim_i file. It's one of two loggers (the other being update) that log both to file and console.
if sys.version < '3.8':  # Python2/3 use different formatting styles
  logger = init_logger('siwim_i', fmt='%(name)-9s %(asctime)s %(levelname)-8s %(message)s', to_console=True)
else:
  logger = init_logger('siwim_i', fmt='%(name)-9s %(asctime)s %(levelname)-8s [%(filename)s: %(lineno)d] %(message)s', to_console=True)

interrupted = False
# Handler for the Interrupt signal with required parameters
def keyboard_interrupt(signum, frame):
  log(logger, 'Caught signal {} (shutdown)'.format(signum))
  global interrupted
  interrupted = True

fd = None
modules = None
module_names = list()

end_i = False

update_host = ""
sites_path = ""

comm_module_name = ""

version = [4, 10, 3]
channel = "stable"

log(logger, 'Running in debug mode.', sev=logging.DEBUG)

try:
  fd = open("conf/conf.xml")
  doc = etree.parse(fd)
  fd.close()
  modules = list(doc.iter("module"))
  update_host = doc.find("update_host").text
  sites_path = doc.find("sites_path").text
  try:
    channel = doc.find("update_channel").text
  except:
    pass
  log(logger, "SiWIM-I v{} starting. Using {} update channel.".format(".".join(str(x) for x in version), channel))
  aggr_cnt = 0
  for module in modules:
    if module.attrib["type"] == "aggr":
      aggr_cnt+=1
      if aggr_cnt > 1:
        raise Exception("Current implementation supports only one (1!) aggregation module!")
    module_name = module.find("name").text
    if module_name in module_names:
      raise Exception("All module names must be unique!")
    module_names.append(module_name)
except:
  log_exception(logger, 'Parsing conf file failed:')
  sys.exit(0)

sites_path = sites_path.replace("\\", "/")
if sites_path[len(sites_path) - 1] == "/":
  sites_path = sites_path[:-1]  
if sites_path == "" or not os.path.exists(sites_path):
  log(logger, "Sites directory doesn't exist. Exiting ...", sev=logging.CRITICAL)
  sys.exit(0)

root_path = sites_path[0:sites_path.find("/sites")]

running = dict()
downstream = dict()  

comm_started = False

for module in modules:
  name = module.find("name").text
  if module.attrib["type"] == "rcv":
    try:
      validity_tag = ""
      try:
        validity_tag = module.find("validity_tag").text
      except:
        pass
      offsets_dict = {}
      try:
        offsets_from_boss = module.find("offset_from_boss").findall("lane")
        num_lane_dict = dict()
        for lane in offsets_from_boss:
          offsets_dict[lane.attrib["num"]] = float(lane.text)
      except:
        pass
      recv_buff = 2048
      try:
        recv_buff = int(module.find("recv_buffer_size").text)
      except:
        pass
      vehicle_end = "0"
      try:
        vehicle_end = module.find("vehicle_end").text
      except:
        pass
      from receive_module import ReceiveModule
      running[name] = ReceiveModule(name, module.find("typ").text, module.find("host").text, int(module.find("port").text), recv_buff, int(module.find("save_original").text), root_path, offsets_dict, validity_tag, vehicle_end)
    except:
      pass
  elif module.attrib["type"] == "aggr":
    try:
      from aggregation_module import AggregationModule
      running[name] = AggregationModule(name, float(module.find("matching_timeout").text), float(module.find("fuzziness").text), int(module.find("save_to_disk").text), module.find("boss_receive").text, root_path)
    except:
      log_exception(logger, 'Got exception adding Aggregation Module:')
  elif module.attrib["type"] == "cam": 
    try:
      init_type = module.find("init_type").text
      params = dict()
      module_children = module.getchildren()
      for module_child in module_children:
        if module_child.tag == "ignore_classes":
          class_lane_dict = dict()
          for lane in module_child.getchildren():
            classes = list()
            for cls in lane.getchildren():
              try:
                classes.append(int(cls.text))
              except:
                pass
            class_lane_dict[int(lane.attrib["num"])] = classes
          params[module_child.tag] = class_lane_dict
        elif module_child.tag == "fixed_offset" or module_child.tag == "cam_focus":
          num_lane_dict = dict()
          for lane in module_child.getchildren():
            num_lane_dict[int(lane.attrib["num"])] = int(lane.text)
          params[module_child.tag] = num_lane_dict
        elif module_child.tag == "lanes":
          lanes = list()
          for lane in module_child.getchildren():
            lanes.append(int(lane.text))
          params[module_child.tag] = lanes
        else:
          params[module_child.tag] = module_child.text
      from camera_module import CameraModule
      running[name] = CameraModule(name, init_type, params, root_path)
    except:
      log_exception(logger, 'Got exception adding Camera Module:')
  elif module.attrib["type"] == "pic":
    try:
      recv_buff = 2048
      try:
        recv_buff = int(module.find("recv_buffer_size").text)
        backlog = int(module.find("backlog").text)
      except:
        backlog = 5
      # Read the list of camera types present on site
      default_cam_types = list(module.iter('cam_type'))
      default_cam_types_text = list()
      if len(default_cam_types) > 0:
        for cam_type in default_cam_types:
          default_cam_types_text.append(cam_type.text)
      else:
        try:
          default_cam_types_text.append(module.find("default_cam_type").text)
        except:
          pass
      default_cam_names = list(module.iter("cam_name"))
      default_cam_names_text = list()
      for default_cam_name in default_cam_names:
        default_cam_names_text.append(default_cam_name.text)
      from picture_module import PictureModule
      running[name] = PictureModule(name, int(module.find("port").text), backlog, recv_buff, root_path, int(module.find("old").text), default_cam_types_text, default_cam_names_text)
    except:
      pass
  elif module.attrib["type"] == "out":
    backlog = 5
    try:
      backlog = int(module.find("backlog").text)
      from output_module import OutputModule
      running[name] = OutputModule(name, int(module.find("port").text), backlog, root_path)
    except:
      pass
  elif module.attrib["type"] == "lrm":
    try:
      url_node = module.find("url")
      protocol_node = module.find("protocol")
      protocol = "http"
      url = "/siwim-s/sys_alarm.php"
      if protocol_node != None:
        protocol = protocol_node.text
      if url_node != None:
        url = url_node.text
      timeouts = dict()
      timeout_nodes = module.find("timeouts").findall("timeout")
      impl_timeout_nodes = module.find("timeouts").findall("impl_interval_timeout")
      pairs = list(zip(timeout_nodes, impl_timeout_nodes))  # In Python2 this is already a list, but casting a list to list does nothing, so ...
      for pair in pairs:
        timeouts[pair[0].attrib["lane"]] = (pair[0].text, pair[1].text)
      from alarm_module import AlarmModule
      running[name] = AlarmModule(name, float(module.find("poll_rate").text), float(module.find("min_u").text), float(module.find("u_tresh").text), float(module.find("min_dd").text), module.find("interval_start").text, module.find("interval_end").text, timeouts, float(module.find("alarm_timeout").text), module.find("host").text, url, protocol, root_path)
    except:
      log_exception(logger, 'Got exception adding Alarm Module:')
  elif module.attrib["type"] == "stat":
    try:
      url_node = module.find("url")
      protocol_node = module.find("protocol")
      protocol = "http"
      url = "/siwim-s/sys_report.php"
      if protocol_node != None:
        protocol = protocol_node.text
      if url_node != None:
        url = url_node.text
      ntp_conf_path = ""
      ntp_conf_node = module.find("ntp_conf_path")
      if ntp_conf_node != None:
        ntp_conf_path = ntp_conf_node.text
      ovpn_path = "/usr/sbin/openvpn"
      if sys.platform == "win32":
        ovpn_path = "c:/program files/openvpn/bin/openvpn.exe"
      ovpn_node = module.find("ovpn_path")
      if ovpn_node != None and ovpn_node.text != str():
        ovpn_path = ovpn_node.text
      save_curr_to_disk = 0
      try:
        save_curr_to_disk = int(module.find("save_curr_to_disk").text) 
      except: 
        log_exception(logger, 'Got exception retrieving save_curr_to_disk:')
      from status_module import StatusModule
      running[name] = StatusModule(name, module.find("host").text, channel, float(module.find("time_interval").text), version, url, protocol, root_path, ntp_conf_path, ovpn_path, save_curr_to_disk)
    except:
      log_exception(logger, 'Got exception adding Status Module:')
  elif module.attrib["type"] == "ctu":
    try:
      from ctu_module import CtuModule
      relevant_for_listeners = "UT"
      try:
        relevant_for_listeners = module.find("relevant_for_listeners").text
      except:
        pass
      running[name] = CtuModule(name, module.find("udp_ip").text, int(module.find("udp_port").text), module.find("serial_port").text, int(module.find("baud").text), float(module.find("interval").text), module.find("vpn_host").text, int(module.find("vpn_port").text), root_path, relevant_for_listeners)
    except:
      log_exception(logger, 'Got exception adding Ctu Module:')
  elif module.attrib["type"] == "can":
    try:
      from can_module import CanModule
      relevant_for_listeners = "T"
      try:
        relevant_for_listeners = module.find("relevant_for_listeners").text
      except:
        pass
      board_ids = list()
      try:
        board_id_nodes = module.find("board_ids").findall("board_id")
        for node in board_id_nodes:
          board_ids.append(int(node.text))
      except:
        board_ids = ["1", "2", "3"]
      running[name] = CanModule(name, module.find("udp_rx_ip").text, int(module.find("udp_rx_port").text), module.find("udp_tx_ip").text, int(module.find("udp_tx_port").text), module.find("udp_e_ip").text, int(module.find("udp_e_port").text), float(module.find("interval").text), root_path, relevant_for_listeners, int(module.find("ap_fwver").text), board_ids)
    except:
      log_exception(logger, 'Got exception adding Can Module:')
      pass
  elif module.attrib["type"] == "hist":
    try:
      from hist_module import HistModule
      relevant_for_listeners = "U"
      try:
        relevant_for_listeners = module.find("relevant_for_listeners").text
      except:
        pass
      running[name] = HistModule(name, module.find("udp_ip").text, int(module.find("udp_port").text), module.find("serial_port").text, int(module.find("baud").text), float(module.find("interval").text), root_path, relevant_for_listeners)
    except:
      log_exception(logger, 'Got exception adding Hist Module:')
  elif module.attrib["type"] == "papago":
    try:
      from papago_module import PapagoModule
      relevant_for_listeners = "U"
      try:
        relevant_for_listeners = module.find("relevant_for_listeners").text
      except:
        pass
      running[name] = PapagoModule(name, module.find("udp_ip").text, int(module.find("udp_port").text), module.find("device_ip").text, float(module.find("interval").text), int(module.find("indexing_start").text), root_path, relevant_for_listeners)
    except:
      log_exception(logger, 'Got exception adding Papago Module:')
  elif module.attrib["type"] == "comm":
    try:
      backlog = 5
      try:
        backlog = int(module.find("backlog").text)
      except:
        pass
      comm_module_name = name
      running[name] = CommunicationModule(name, int(module.find("port").text), backlog, int(module.find("recv_buffer_size").text), root_path)
      comm_started = True
    except:
      log_exception(logger, 'Got exception adding Communication Module:')
  try:
    running[name].set_sites_path(sites_path)
    dm_list = list(module.iter("downstream_module"))
    downstream[name] = list()
    for dm in dm_list:
      downstream[name].append(dm.text)
  except:
    log_exception(logger, 'Got exception retrieving downstream_module:')
  
if not comm_started:
  name = "comm_default"
  comm_module_name = name
  downstream[name] = list()
  running[name] = CommunicationModule(name, 8173, 5, 2048, root_path)

#print downstream
#a hacky hack, because a alarm module NEEDS to be downstream of a ctu module in order to supply voltage alarms
#this may need to be removed some day, as it imposes some rules, that may be false in the future
#for ctu_module_name in ctu_module_names:
#  for alarm_module_name in alarm_module_names:
#    if ctu_module_name not in downstream:
#      downstream[ctu_module_name] = list()
#    if alarm_module_name not in downstream[ctu_module_name]:
#      downstream[ctu_module_name].append(alarm_module_name)
#print downstream
      
for key, module in running.items(): 
  log(logger, "Starting " + key + " module ...")
  for downster in downstream[key]:
    try:
      module.add_downstream_module(running[downster])
    except:
      pass
  threading.Thread(target=module.run).start()

# Whenever Ctrl+C is pressed, sig_handler, which sets interrupted variable, will be called.
signal.signal(signal.SIGINT, keyboard_interrupt)
day = 0
if print_memory_usage:  # The flag is set by passing a parameter
  try:
    from guppy import hpy  # This library causes a crash if used with win32com.client (which we don't use here)
    print_count = 60
  except:
    log(logger, 'Failed to import guppy. Please install guppy (Python2) or guppy3 (Python3) to enable -m flag.', sev=logging.WARNING)
    print_memory_usage = False
while True:
  if print_memory_usage:
    if print_count == 60:
      print(hpy().heap())
      print_count = 0
    print_count += 1
  for key, module in running.items():
    if not module.is_alive():
      try:
        log(logger, "Restarting " + key + " module ...")
        threading.Thread(target=module.run).start()
      except:
        log_exception(logger, 'Got exception while trying to restart a module:')
  end_i = False
  loc = ""
  new_ver_list = [0, 0, 0]
  try:
    if day != datetime.datetime.now().day and update_host != "":
      log(logger, 'Host: {0}, Channel: {1}'.format(update_host, channel), sev=logging.DEBUG)
      day = datetime.datetime.now().day
      resp = requests.post('http://{0}/check.php?sw=siwim-i&ch={1}'.format(update_host, channel), timeout=10).text
      resp_parts = resp.split()
      for part in resp_parts:
        if part[0:4] == "ver=":
          new_ver_list = part[4:].split(".")
          for i in range(len(new_ver_list)):
            new_ver_list[i] = int(new_ver_list[i])
        if part[0:4] == "loc=":
          loc = part[4:]
          log(logger, 'Update loc: {0}'.format(loc), sev=logging.DEBUG)
      loc = loc[loc.find("/"):]
      for i in range(len(version)):
        if version[i] < new_ver_list[i]:
          end_i = True
          break
        elif version[i] > new_ver_list[i]:
          break
  except:
    log(logger, 'No update found.', sev=logging.DEBUG)
  if end_i:
    log(logger, "Updating SiWIM-I to " + ".".join(str(x) for x in new_ver_list))
    end_i = True
    if not os.path.exists("update"):
      os.mkdir("update/")
    file = requests.get("http://{0}{1}".format(update_host, loc), stream=True)
    with open('update/update.zip', 'wb') as f:
      for data in file.iter_content(chunk_size=1024):
        f.write(data)
    #delete files to make sure update directory is ready for use
    up_files = os.listdir("update")
    for up_file in up_files:
      if up_file != "update.zip" and up_file != "i_updater.py":
        os.remove("update/" + up_file)
    zip_ref = zipfile.ZipFile("update/update.zip", 'r')
    zip_ref.extractall("update")
    zip_ref.close()
    os.remove("update/update.zip")                               
    subprocess.Popen([sys.executable, "update/i_updater.py", str(os.getpid())])
  elif running[comm_module_name].are_we_resetting_1():
    log(logger, "Resetting I due to a frontend message ...")
    end_i = True
    if sys.platform == "win32":
      subprocess.Popen([sys.executable, "update/i_updater.py", str(os.getpid())])
  elif interrupted:
    end_i = True
    log(logger, 'Received signal interrupt.')
  if end_i:
    log(logger, 'Shutting down.')
    for key1, module1 in running.items():
      module1.set_end()
    while True:
      all_dead = True
      for key1, module1 in running.items():
        if module1.is_alive():
          log(logger, 'Waiting for {0} to die'.format(key1), sev=logging.DEBUG)
          all_dead = False
          break
      if all_dead:
        log(logger, 'All threads have closed.')
        break
      try:
        time.sleep(0.5)
      except:
        log(logger, 'Sleep failed?')
    log(logger, 'Calling os._exit(0) on siwim_i.py', sev=logging.DEBUG)
    os._exit(0)
  try:
    time.sleep(1)
  except:
    log(logger, 'Exception during sleep (keyboard interrupt, maybe)')
  